/* 
   Authors: Loic Besnard, Patricia Bournai, Thierry Gautier, Paul Le Guernic.
   Copyright (C) by INRIA-CNRS - developed by Espresso project - GPL v2 License - 2010.

   Contact: polychronycontact@listes.irisa.fr 

   This file is part of Polychrony Signal Toolbox (http://www.irisa.fr/espresso/polychrony).

   Polychrony  Signal Toolbox is free software; you can redistribute it and/or modify it under
   the terms of the GNU General Public License Version 2 as published by the Free
   Software Foundation.

   Polychrony Signal Toolbox is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License Version 2
   for more details.

   You should have received a copy of the GNU General Public License Version 2
   along with Polychrony Signal Toolbox; see the file COPYING. If not, consult the url 
    http://www.gnu.org/licenses/gpl-2.0.html, or write to the Free
   Software Foundation, 51 Franklin Street, Fifth Floor, Boston, MA
   02110-1301, USA. 
*/

#include <cstdio>
#include <cstdlib>
#include "pK_scheduler.h"

TaskFifo::TaskFifo (int task_number, Task ** task_list)
{
  _TasksNumber = task_number;
  _TaskList = task_list;
  _TaskFifo = (Task **) malloc (_TasksNumber * sizeof (Task *));
}

TaskFifo::~TaskFifo ()
{
  free ((char *) _TaskFifo);
}

// return the number of tasks
int
TaskFifo::GetTasksNumber ()
{
  return _TasksNumber;
}

// return a task
Task *
TaskFifo::GetTask (int n)
{
  return _TaskList[n];
}

// initializes the fifo
void
TaskFifo::ReInit ()
{
  _FillIndex = 0;
  _ExtractIndex = 0;
  _InsideNumber = 0;
}

// pushes a task into the fifo
void
TaskFifo::Push (Task * t)
{
  _TaskFifo[_FillIndex] = t;
  _FillIndex = (_FillIndex + 1) % _TasksNumber;
  _InsideNumber++;
}

// pops a task from the fifo (NULL when empty)
Task *
TaskFifo::Pop ()
{
  Task *t;

  if (!IsEmpty ())
    {
      t = _TaskFifo[_ExtractIndex];
      _ExtractIndex = (_ExtractIndex + 1) % _TasksNumber;
      _InsideNumber--;
      return t;
    }
  return NULL;
}

// return 0 when the fifo is empty, 1 otherwise
int
TaskFifo::IsEmpty ()
{
  return (_InsideNumber == 0);
}




Scheduler::Scheduler (int task_number, Task ** task_list, GetNewInputFunc GetNewInput, UpdateInputFunc UpdateInput, SendMessFunc SendMess, BroadCastMessFunc BroadCastMess, char *Name):TaskFifo (task_number, task_list),
_GetNewInput (GetNewInput),
_UpdateInput (UpdateInput), _SendMess (SendMess), _BroadCastMess (BroadCastMess), _Verbose (0), _Name (Name), _InstantNumber (0), _TraceFile (NULL)
{
#ifdef VERBOSE
  SetVerbose ();
#endif
}

void
Scheduler::Initialize ()
{
  int i;
  for (i = 0; i < GetTasksNumber (); i++)
    GetTask (i)->Initialize ();
}

void
Scheduler::SetVerbose ()
{
  _Verbose = 1;
}

int
Scheduler::IsVerbose ()
{
  return _Verbose;
}

void
Scheduler::Trace (const char *string)
{
  char str[1024];
  if (IsVerbose ())
    {
      if (_TraceFile == NULL)
	{
	  sprintf (str, "%s.trace", _Name);
	  _TraceFile = fopen (str, "wt");
	};
      // fprintf(_TraceFile,"%s : %s\n",_Name,string); 
      fprintf (_TraceFile, "%s\n", string);
      fflush (_TraceFile);
    }
}

void
Scheduler::TraceInt (const char *string, int n)
{
  char str[1024];

  if (IsVerbose ())
    {
      sprintf (str, string, n);
      Trace (str);
    }
}

void
Scheduler::TraceStr (const char *string, const char *s)
{
  char str[1024];

  if (IsVerbose ())
    {
      sprintf (str, string, s);
      Trace (str);
    }
}

void
Scheduler::TraceStrInt (const char *string, const char *s, int n)
{
  char str[1024];

  if (IsVerbose ())
    {
      sprintf (str, string, s, n);
      Trace (str);
    }
}



void
Scheduler::NewInstant ()
{
  char str[1024];

  _InstantNumber++;
  if (IsVerbose ())
    {
      sprintf (str, " ==================================== instant %d.", _InstantNumber);
      Trace (str);
    }
}

MasterScheduler::MasterScheduler (int task_number, Task ** task_list, GetNewInputFunc GetNewInput, UpdateInputFunc UpdateInput, SendMessFunc SendMess,
				  BroadCastMessFunc BroadCastMess, int slaves_number, char *Name):
Scheduler (task_number, task_list, GetNewInput, UpdateInput, SendMess, BroadCastMess, Name),
_SlavesNumber (slaves_number)
{
}

int
MasterScheduler::Iterate ()
{
  Task *t;
  SchedulerState status;
  int i;

  NewInstant ();
  ReInit ();
  status = __NORMAL__;
  do
    {
      if (!IsEmpty ())
	{
	  //Trace("IsEmpty ==> false.");
	  t = Pop ();
	  //Trace("sortie t->Pop.");
	  t->Select ();
	  //Trace("Sortie t->Select.");
	  t->Iterate ();
	  //Trace("t->Iterate ==> true.");
	  t->Terminate ();
	  _Weight--;
	}
      else
	{
	  //Trace("IsEmpty ==> true.");
	  if (!Scan ())
	    {
	      Trace ("------> __ERROR__ received.");
	      /* __ERROR__ message received */
	      status = __ERROR__;
	    }
	};
    }
  while ((status == __NORMAL__) && ((_Weight != 0) || (_SyncWeight != 0)));
  switch (status)
    {
    case __NORMAL__:
      return 1;
    case __ERROR__:
      Trace ("------> broadcast __END__.");
      _BroadCastMess (__END_STAMP__);
      /* wait for _SlavesNumber __END__ messages */
      for (i = 0; i < _SlavesNumber; i++)
	ScanEnd ();
      Trace ("exiting...");
      return 0;
    }
  return -1;			// unused
}

void
MasterScheduler::ReInit ()
{
  int i;

  TaskFifo::ReInit ();
  for (i = 0; i < GetTasksNumber (); i++)
    GetTask (i)->ReInit ();
  _Weight = GetTasksNumber ();
  _SyncWeight = _SlavesNumber;
}

/* scan data; return false when receiving an __ERROR__ message */
int
MasterScheduler::Scan ()
{
  int stamp;
  char *data;

  if (!_GetNewInput (&stamp, &data))
    TraceAndStop ();
  switch (stamp)
    {
    case __ERROR_STAMP__:
      Trace ("------> __ERROR__ received.");
      return 0;
    case __SYNC_STAMP__:
      Trace ("------> __SYNC__ received.");
      _SyncWeight--;
      break;
    default:
      _UpdateInput (stamp, data);
    }
  return 1;
}

void
MasterScheduler::ScanEnd ()
{
  int stamp;
  char *data;

  Trace ("waiting for __END__.");
  do
    _GetNewInput (&stamp, &data);
  while (stamp != __END_STAMP__);
}

/* function called on timeout */
void
MasterScheduler::TraceAndStop ()
{
  Trace ("-- timeout -- timeout-- timeout --");
  exit (1);
}


SlaveScheduler::SlaveScheduler (int task_number, Task ** task_list, GetNewInputFunc GetNewInput, UpdateInputFunc UpdateInput, SendMessFunc SendMess,
				BroadCastMessFunc BroadCastMess, int master, char *Name):
Scheduler (task_number, task_list, GetNewInput, UpdateInput, SendMess, BroadCastMess, Name),
_Master (master)
{
}

int
SlaveScheduler::Iterate ()
{
  Task *t;
  SchedulerState status;

  NewInstant ();
  ReInit ();
  status = __NORMAL__;
  do
    {
      if (!IsEmpty ())
	{
	  t = Pop ();
	  t->Select ();
	  t->Iterate ();
	  t->Terminate ();
	  _Weight--;
	}
      else
	{
	  if (!Scan ())
	    /* end message received */
	    status = __END__;
	};
    }
  while ((status == __NORMAL__) && (_Weight != 0));
  switch (status)
    {
    case __NORMAL__:
      /* synchronization */
      Trace ("------> sending __SYNC__.");
      _SendMess (_Master, __SYNC_STAMP__);
      return 1;
    case __ERROR__:
      /* wait for an __END__ message */
      Trace ("waiting for __END__.");
      while (Scan ());
    case __END__:
      /* send last _END__ message */
      Trace ("------> sending __END__.");
      _SendMess (_Master, __END_STAMP__);
      Trace ("exiting...");
      return 0;
    }
  return -1;			// unused
}

void
SlaveScheduler::ReInit ()
{
  int i;

  TaskFifo::ReInit ();
  for (i = 0; i < GetTasksNumber (); i++)
    GetTask (i)->ReInit ();
  _Weight = GetTasksNumber ();
}

/* scan data; return false when receiving an __END__ message */
int
SlaveScheduler::Scan ()
{
  int stamp;
  char *data;

  if (!_GetNewInput (&stamp, &data))
    TraceAndStop ();
  switch (stamp)
    {
    case __END_STAMP__:
      Trace ("------> __END__ received.");
      return 0;
    default:
      _UpdateInput (stamp, data);
    }
  return 1;
}

/* function called on timeout */
void
SlaveScheduler::TraceAndStop ()
{
  Task *task;
  int i, j;

  Trace ("-- timeout -- timeout-- timeout --");
  TraceInt ("Tasks not fired : %d", _Weight);
  if (_Weight != 0)
    {
      for (i = 0; i < GetTasksNumber (); i++)
	{
	  task = GetTask (i);
	  if (task->IsNotBegan ())
	    {
	      TraceStrInt ("    %s (%d)", task->GetName (), task->GetCounter ());
	      if (task->GetSuccNumber () != 0)
		{
		  for (j = 0; j < task->GetSuccNumber (); j++)
		    TraceStr ("        --> %s", task->GetSucc (j)->GetName ());
		}
	    }
	}
    }
  exit (1);
}

Task::Task (Scheduler * Sched, IterFunc Iter,
	    //      InitFunc Init, 
	    int PredNumber, int SuccNumber, Task * Succ[], char *Name):
_Sched (Sched),
_PredNumber (PredNumber),
_SuccNumber (SuccNumber),
_Succ (Succ),
_Iter (Iter),
  //  _Init ( Init ), 
_State (_terminated),
_Name (Name)
{
}

Task::Task (Scheduler * Sched, IterFunc Iter,
	    //      InitFunc Init, 
	    int PredNumber, char *Name):
_Sched (Sched),
_PredNumber (PredNumber),
_SuccNumber (0),
_Succ (NULL),
_Iter (Iter),
  //  _Init ( Init ), 
_State (_terminated),
_Name (Name)
{
}

void
Task::Iterate ()
{
  Trace ("Iterate.");
  _Iter ();
}

void
Task::Initialize ()
{				/* _Init ( ) ; */
}

void
Task::ReInit ()
{
  _Counter = _PredNumber;
  if (_Counter == 0)
    Start ();
  else
    _State = _not_began;
}

void
Task::Start ()
{
  //Trace("Start."); 
  _Sched->Push (this);
  SetReady ();
}

void
Task::SetReady ()
{
  //Trace("SetReady."); 
  _State = _ready;
}

void
Task::Select ()
{
  Trace ("Select.");
  _State = _running;
}

void
Task::Terminate ()
{
  int i;

  Trace ("Terminate.");
  for (i = 0; i < _SuccNumber; i++)
    _Succ[i]->DecCounter ();
  _State = _terminated;
}

void
Task::DecCounter ()
{
  Trace ("DecCounter.");
  _Counter--;
  if (_Counter == 0)
    Start ();
}

void
Task::Trace (const char *string)
{
  char str[1024];
  if (_Sched->IsVerbose ())
    {
      sprintf (str, "task %s : %s", _Name, string);
      _Sched->Trace (str);
    }
}

void
Task::TraceBool (const char *string, int b)
{
  char str[1024];

  sprintf (str, string, b);
  Trace (str);
}

int
Task::IsNotBegan ()
{
  return _State == _not_began;
}

int
Task::GetCounter ()
{
  return _Counter;
}

char *
Task::GetName ()
{
  return _Name;
}

int
Task::GetSuccNumber ()
{
  return _SuccNumber;
}

Task *
Task::GetSucc (int n)
{
  return _Succ[n];
}
